let PIXI;
let Game = function (options) {
  this.options = Object.assign({}, options);

  let padding = Object.assign({
    left: 0,
    top: 0,
    right: 0,
    bottom: 0
  }, this.options.gameSource.padding);
  this.scale = this.options.$game.fObj.scale;
  this.padding = {
    left: padding.left * this.scale,
    right: padding.right * this.scale,
    top: padding.top * this.scale,
    bottom: padding.bottom * this.scale,
  };
  PIXI = this.options.$game.fObj.PIXI;
  this.width = this.options.$game.fObj.size.windowWidth;
  this.height = this.options.$game.fObj.size.windowHeight;

  // 1.创建pixi需要的对象 容器
  // 精灵容器
  this.spriteContainer = new PIXI.Container();
  const graphics = new PIXI.Graphics();
  graphics.beginFill(0xFFFFFF);
  let circleObj = Object.assign({
    x: this.width / 2 / this.scale, y: this.height / 2 / this.scale, r: this.width / 2 / this.scale
  }, this.options.gameSource.mask);
  // 四个边的边界
  this.BLeft = circleObj.x * this.scale - circleObj.r * this.scale;
  this.BRight = circleObj.x * this.scale + circleObj.r * this.scale;
  this.BTop = circleObj.y * this.scale - circleObj.r * this.scale;
  this.BBottom = circleObj.y * this.scale + circleObj.r * this.scale;

  this.clickInfo = { x: circleObj.x * this.scale, y: circleObj.y * this.scale };
  graphics.drawCircle(circleObj.x * this.scale, circleObj.y * this.scale, circleObj.r * this.scale);
  graphics.endFill();
  this.spriteContainer.mask = graphics;
  this.options.$game.fObj.application.stage.addChild(this.spriteContainer);
  // 操作元素容器 (只有一个元素)
  this.playerContainer = new PIXI.Container();
  this.options.$game.fObj.application.stage.addChild(this.playerContainer);
  // 动画容器
  this.animateContainer = new PIXI.Container();
  this.options.$game.fObj.application.stage.addChild(this.animateContainer);

  // 2.遍历图片资源 并 去重
  let imgArr = [];
  options.gameSource = JSON.parse(JSON.stringify(options.gameSource));
  this.traverse(options.gameSource, function (k, v) {
    if (k == "src" && !!v && !imgArr.includes(v) && !PIXI.utils.TextureCache[v]) {
      imgArr.push(v);
    }
  })
  if (!!options.touchObj) {
    options.touchObj = JSON.parse(JSON.stringify(options.touchObj));
    this.traverse(options.touchObj, function (k, v) {
      if (k == "src" && !!v && !imgArr.includes(v) && !PIXI.utils.TextureCache[v]) {
        imgArr.push(v);
      }
    })
  }

  // 定时更新
  this.ticker = PIXI.ticker.shared;
  this.ticker.autoStart = false;
  this.ticker.addOnce(() => { });

  // 3.加载图片资源
  if (imgArr.length > 0) {
    let loader = new PIXI.loaders.Loader();
    imgArr.forEach(item => {
      loader.add(this.uniId(), item);
    })
    this.pixiLoadFun(loader);
  } else {
    this.init({});
    this.options.initDone();
  }

  this.ticker.add(() => this.draw());
}
Game.prototype = Object.assign(Game.prototype, {
  pixiLoadFun: function (loader) {
    let time1 = new Date().getTime();
    loader.on("progress", e => {
      this.options.progressFun(e.progress);
    })
    loader.on("error", e => {
      this.options.errorFun(e);
    })
    loader.load(() => {
      // 资源加载完成
      time1 = "";
      this.init({});
      this.options.initDone();
    });
    setTimeout(() => {
      // 5秒还没加载完，重新加载一遍
      if (!!time1) {
        console.log("重新加载资源...")
        this.pixiLoadFun(loader);
      }
    }, 5000)
  },
  uniId: function () {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
      var r = Math.random() * 16 | 0, v = c == 'x' ? r : (r & 0x3 | 0x8);
      return v.toString(16);
    });
  },
  random: function (options) {
    let opt = Object.assign({ min: 0, max: 0, isFloat: false, noNum: [] }, options);
    let tempNum;
    do {
      let num = Math.random() * (opt.max - opt.min) + opt.min;
      tempNum = !!opt.isFloat ? num : parseInt(num);
    } while (opt.noNum.includes(tempNum));
    return tempNum;
  },
  isArray: function (obj) {
    return obj instanceof Array;
  },
  isObject: function (obj) {
    // 判断数组对象
    if (!(obj instanceof Array) && (obj instanceof Object)) {
      return true;
    }
    return false;
  },
  traverse: function (obj, callback) {
    // 遍历json
    for (let name in obj) {
      // 如果是json对象
      if (this.isObject(obj[name])) {
        this.traverse(obj[name], callback);
      } else if (this.isArray(obj[name])) {
        // 数组对象
        obj[name].forEach(item => {
          this.traverse(item, callback);
        })
      } else {
        callback(name, obj[name]);
      }
    }
  },
  tickerStart() {
    let _this = this;
    if (!!_this.req) { return false; }
    this.tickerStopStatus = false;

    this.ticker.start();
    this.req = true;
  },
  tickerStop() {
    this.iStart = false;
    this.tickerStopStatus = true;

    this.ticker.stop();
    this.req = null;
  },
  frame: function (img, x, y) {
    // 返回texture
    let source = img.src, width = img.width, height = img.height;

    let texture, imageFrame;
    if (typeof source === "string") {
      if (PIXI.utils.TextureCache[source]) {
        texture = new PIXI.Texture(PIXI.utils.TextureCache[source]);
      }
    }
    if (!texture) { console.log(`Please load the ${source} texture into the cache.`); } else {
      imageFrame = new PIXI.Rectangle(x, y, width, height);
      texture.frame = imageFrame;
      return texture;
    }
  },
  renderSprite: function (tImg, ops) {
    let sprite;
    // 没有唯一标识，说明第一次生成
    if (tImg.type == "animate") {
      // 动画
      let textures = [];
      tImg.srcArr.forEach(img => {
        textures.push(this.frame(img, 0, 0));
      })
      sprite = new PIXI.extras.AnimatedSprite(textures);
      sprite.animationSpeed = tImg.boomSpeed;
      sprite.loop = false;
      sprite.onComplete = function () {
        if (!!tImg.maxAnimate) {
          // 执行回调
          try {
            tImg.callback(sprite);
          } catch (e) { }
        }
      }
    } else {
      sprite = new PIXI.Sprite();

      let options = Object.assign({ x: 0, y: 0 }, ops);
      sprite.texture = this.frame(tImg, options.x, options.y);
    }

    sprite.width = tImg.width * this.scale;
    sprite.height = tImg.height * this.scale;

    sprite.x = (tImg.x || 0);
    sprite.y = (tImg.y || 0);
    sprite.imgObj = tImg;
    if (!tImg.noAnchor) {
      sprite.anchor.set(0.5, 0.5);
    }

    if (!tImg.noScale) {
      sprite.scale.set(tImg.scale || 1, tImg.scale || 1);
    }

    sprite.uniId = this.uniId();

    sprite.type = tImg.type || "";
    let name = "", con = "";
    if (tImg.type == "animate") {
      name = "animate";
      con = "animate";

      sprite.stop()
      sprite.gotoAndPlay(0)
    } else if (tImg.type == "mask") {
      name = "no";
      con = "player";
      this.maskSprite = sprite;
    } else {
      con = "sprite";
    }
    if (name != "no") {
      this[`${name}spriteArr`].push(sprite);
    }
    this[`${con}Container`].addChild(sprite);
    if (tImg.type == "mask") {
      this.playerContainer.setChildIndex(sprite, 0);
    }

    return sprite;
  },
  // 添加精灵
  addBlock: function () {
    // 当前精灵个数
    let energyNum = 0, barrierNum = 0;
    // 可显示的总精灵个数
    let totalEnergyNum = this.options.gameSource.energyNum, totalBarrierNum = this.options.gameSource.barrierNum;

    let total = this.spriteArr.length;
    let i = 0;
    while (i < total) {
      let sprite = this.spriteArr[i++];
      if (sprite.visible) {
        if (sprite.type == "energy") {
          energyNum++;
        } else if (sprite.type == "barrier") {
          barrierNum++;
        }
      }
    };
    if (energyNum < totalEnergyNum) {
      // 生成能量精灵
      if (this.options.gameSource.energyArr) {
        this.renderBlock("energy", this.options.gameSource.energyArr);
      }
    } else {
      this.firstRender = true;
    }
    if (barrierNum < totalBarrierNum) {
      // 生成障碍物精灵
      if (this.options.gameSource.barrierArr) {
        this.renderBlock("barrier", this.options.gameSource.barrierArr);
      }
    }
  },
  renderBlock: function (type, imgArr) {
    if (this.firstRender) {
      let curTime = new Date().getTime();
      if (!!this[`${type}Time`]) {
        if (curTime - this[`${type}Time`] <= 1000) {
          // 相同元素间隔时间
          return false;
        }
      }
      this[`${type}Time`] = curTime;
    }
    let renderArr = [];//需要生成的精灵数组
    let totalProbability = 0;
    imgArr.forEach(item => {
      item.probability = item.probability || 1;
      item.max = totalProbability + item.probability
      totalProbability = item.max;
      renderArr.push(Object.assign({}, item));
    });

    // 此次生成的元素
    let defaultImg;
    let tempIndex = this.random({ max: totalProbability });
    renderArr.forEach(item => {
      if (!defaultImg && item.max > tempIndex) {
        defaultImg = Object.assign({
          type: type,
        }, item);
      }
    })

    if (!!defaultImg) {
      // x轴需要随机展示方向
      let directionX = this.random({ max: 10 }) < 5 ? -1 : 1;
      defaultImg.speedX = this.random({ min: this.options.gameSource.speedRang.x.min, max: this.options.gameSource.speedRang.x.max, isFloat: true }) * directionX;
      defaultImg.speedY = this.random({ min: this.options.gameSource.speedRang.y.min, max: this.options.gameSource.speedRang.y.max, isFloat: true });
      defaultImg.scale = this.random({ min: this.options.gameSource.scaleRang.min, max: this.options.gameSource.scaleRang.max, isFloat: true });
      let yiban = defaultImg.width * this.scale * defaultImg.scale / 2;

      let tempD = Object.assign({}, defaultImg);
      tempD.width = defaultImg.width * this.scale;
      tempD.x = this.random({ min: this.BLeft + yiban, max: this.BRight - yiban, isFloat: true });

      if (this.firstRender) {
        tempD.y = -yiban;
      } else {
        let count = 0;
        do {
          tempD.y = this.random({ min: this.BTop + yiban, max: this.BBottom - yiban, isFloat: true });
          count++;
        } while (this.checkCollision(tempD) && count < 10);
      }

      defaultImg.x = tempD.x;
      defaultImg.y = tempD.y;

      if (!this.checkCollision(defaultImg)) {
        this.renderSprite(defaultImg);
      }
    }
  },
  // 获取了能量
  updateFun: function (sprite, scoreObj) {
    this.totalScore += scoreObj.score;
    if (!this.options.gameSource.noVibrateShort) {
      my.vibrateShort({
        success: () => { }
      });
    }
    if (!!sprite.imgObj.activeImg) {
      // 播放动画
      let w = sprite.imgObj.activeImg[0].width, h = sprite.imgObj.activeImg[0].height;
      let animateObj = Object.assign({}, {
        type: "animate",
        maxAnimate: true,
        srcArr: sprite.imgObj.activeImg,
        boomSpeed: sprite.imgObj.activeSpeed,
        scale: sprite.imgObj.scale,
        width: w,
        height: h,
        x: sprite.x,
        y: sprite.y,
        callback: (spri) => {
          setTimeout(() => {
            try {
              spri.visible = false;
              this.options.updateFun({ sprite: sprite, curScore: scoreObj.score, totalScore: this.totalScore });
            } catch (e) { }
          }, (sprite.imgObj.fadeTime || 0.001) * 1000);
        }
      });
      this.renderSprite(animateObj);
    } else {
      try {
        this.options.updateFun({ sprite: sprite, curScore: scoreObj.score, totalScore: this.totalScore });
      } catch (e) { }
    }
  },
  checkWollXCollision: function (sprite) {
    let yiban = sprite.width / 2;
    let minX = this.BLeft + yiban, maxX = this.BRight - yiban;
    let bl = false;
    if (sprite.x <= minX) {
      bl = true;
      sprite.imgObj.speedX = Math.abs(sprite.imgObj.speedX);
    }
    if (sprite.x >= maxX) {
      bl = true;
      sprite.imgObj.speedX = -Math.abs(sprite.imgObj.speedX);
    }
    return bl;
  },
  checkWollYCollision: function (sprite) {
    let yiban = sprite.width / 2;
    let minY = this.BTop + yiban, maxY = this.BBottom - yiban;
    let bl = false;
    if (sprite.y <= minY) {
      bl = true;
      if (!!sprite.changeDirection) {
        sprite.imgObj.speedY = Math.abs(sprite.imgObj.speedY);
      }
    }
    if (sprite.y >= maxY) {
      bl = true;
      if (!!sprite.changeDirection) {
        sprite.imgObj.speedY = -Math.abs(sprite.imgObj.speedY);
      }
    }
    if (sprite.y > minY && !sprite.changeDirection) {
      sprite.changeDirection = true;
    }
    return bl;
  },
  // 检测碰撞，改变方向
  checkCollision: function (checkSprite) {
    let total = this.spriteArr.length;
    let i = 0;
    let bl = false;//默认没有碰撞
    while (i < total) {
      let sprite = this.spriteArr[i++];
      if (sprite.visible && checkSprite.uniId != sprite.uniId) {
        let X = Math.pow((sprite.x - checkSprite.x), 2);
        let Y = Math.pow((sprite.y - checkSprite.y), 2);
        let v = Math.sqrt(X + Y, 2);
        if (v <= sprite.width + checkSprite.width >> 1) {
          // 碰撞了
          bl = true;
          i = total;
          if (!!checkSprite.imgObj) {
            this.changeSpeed(checkSprite, sprite);
          }
        }
      }
    }

    return bl;
  },
  changeSpeed: function (sprite1, sprite2) {
    let pos1 = { x: sprite1.x, y: sprite1.y, r: sprite1.width >> 1 };
    let pos2 = { x: sprite2.x, y: sprite2.y, r: sprite2.width >> 1 };
    //判断传过来的小球对象，相对于碰撞小球的哪个方位
    if (pos1.x < pos2.x) {
      if (pos1.y < pos2.y) {
        //小球对象在被碰小球的左上角
        pos1.yflag = false;
        pos1.xflag = false;
      } else if (pos1.y > pos2.y) {
        //小球对象在被碰小球的左下角
        pos1.xflag = false;
        pos1.yflag = true;
      } else {
        //小球对象在被撞小球的正左方
        pos1.xflag = false;
      }
    } else if (pos1.x > pos2.x) {
      if (pos1.y < pos2.y) {
        //小球对象在被碰撞小球的右上方
        pos1.yflag = false;
        pos1.xflag = true;
      } else if (pos1.y > pos2.y) {
        //小球对象在被碰撞小球的右下方
        pos1.xflag = true;
        pos1.yflag = true;
      } else {
        //小球对象在被撞小球的正右方
        pos1.xflag = true;
      }
    } else if (pos1.y > pos2.y) {
      //小球对象在被撞小球的正下方
      pos1.yflag = true;
    } else if (pos1.y < pos2.y) {
      //小球对象在被撞小球的正上方
      pos1.yflag = false;
    }

    if (pos1.xflag) {
      sprite1.imgObj.speedX = Math.abs(sprite1.imgObj.speedX);
    } else {
      sprite1.imgObj.speedX = -Math.abs(sprite1.imgObj.speedX);
    }
    if (pos1.yflag) {
      sprite1.imgObj.speedY = Math.abs(sprite1.imgObj.speedY);
    } else {
      sprite1.imgObj.speedY = -Math.abs(sprite1.imgObj.speedY);
    }
  },
  draw: function () {
    // 1.判断精灵个数并添加
    this.addBlock();
    // 2.精灵根据自己的速度移动，并检测是否有碰撞
    let obj = null;
    let total = this.spriteArr.length;
    let scoreObj;
    let i = 0;
    while (i < total) {
      let sprite = this.spriteArr[i++];
      if (sprite.visible) {
        this.checkWollXCollision(sprite);
        this.checkWollYCollision(sprite);

        this.checkCollision(sprite);

        let imgObj = sprite.imgObj;
        let yiban = sprite.width / 2;
        let minX = this.BLeft + yiban, maxX = this.BRight - yiban;
        let minY = this.BTop + yiban, maxY = this.BBottom - yiban;
        let x = sprite.x + imgObj.speedX;
        let y = sprite.y + imgObj.speedY;
        if (x <= minX) {
          x = minX;
        }
        if (x >= maxX) {
          x = maxX;
        }
        if (y <= minY) {
          y = minY;
        }
        if (y >= maxY) {
          y = maxY;
        }
        sprite.x = x;
        sprite.y = y;

        if (this.shootStatus) {
          // 射击
          let x = sprite.x;
          let y = sprite.y;
          if (!!this.clickInfo) {
            // 点击坐标
            let X2 = Math.pow(this.clickInfo.x - x, 2);
            let Y2 = Math.pow(this.clickInfo.y - y, 2);
            let v = Math.sqrt(X2 + Y2, 2);
            let shootR = this.options.gameSource.mask.shootR * this.scale;

            if (v <= (sprite.width >> 1) + shootR) {
              // 射击中了能量元素
              if (imgObj.type == "energy") {
                // 判断击中的百分比
                let per = v / ((sprite.width >> 1) + shootR);
                per = Math.min(per * 100, 100);
                imgObj.score.forEach(item => {
                  if (item.min < per && per <= item.max) {
                    scoreObj = item;
                  }
                })
                sprite.visible = false;
                obj = sprite;
              }
              if (imgObj.type == "barrier") {
                // 击中了障碍物
                sprite.visible = false;
              }
            }
          }
        }
      }
    }
    //3.如果有击中元素，则回调更新
    if (!!obj) {
      this.updateFun(obj, scoreObj);
    } else {
      if (this.shootStatus) {
        // 没有击中也回调返回
        this.options.updateFun({ curScore: 0, totalScore: this.totalScore });
      }
    }
    this.shootStatus = false;
  },
  hideSpriteArrFun(name, ops) {
    let options = Object.assign({ childs: [] }, ops);
    if (!!this[name]) {
      this[name].forEach(sprite => {
        if (!!sprite.visible) {
          sprite.visible = false;
        }
        options.childs.forEach(t => {
          if (!!sprite[t] && !!sprite[t].visible) {
            sprite[t].visible = false;
          }
        })
      })
    }
  },
  init: function (defaultData = {}) {
    this.hideSpriteArrFun("animatespriteArr");
    this.hideSpriteArrFun("spriteArr");
    // 障碍物
    this.spriteArr = defaultData.spriteArr || [];// 存放精灵数组
    this.animatespriteArr = defaultData.animatespriteArr || [];// 存放动画精灵数组
    this.shootStatus = false;
    this.firstRender = false;

    this.totalScore = defaultData.totalScore || 0;

    // 蒙层图片
    if (!!this.options.gameSource.mask.src && !this.maskSprite) {
      let maskImg = Object.assign({
        type: "mask",
        noAnchor: true,
        noScale: true
      }, this.options.gameSource.mask);
      maskImg.x = 0;
      maskImg.y = 0;
      this.renderSprite(maskImg);
    }

    my.t = this;

    this.draw();
  }
})
// 射击
Game.prototype.shoot = function () {
  if (this.iStart != "start") {
    return false;
  }
  this.shootStatus = true;
}
// 游戏继续
Game.prototype.continue = function () {
  if (this.iStart != "pause" && this.iStart != "over") {
    return false;
  }
  // 继续游戏，先隐藏之前碰撞到的障碍物
  this.hideSpriteArrFun("animatespriteArr");
  if (!!this.overSprite) {
    this.overSprite.visible = false;
  }
  if (!!this.pauseTime && !!this.startTime) {
    let curTime = new Date().getTime();
    this.startTime += (curTime - this.pauseTime);
    this.pauseTime = 0;
  }
  this.iStart = false;
  this.startGame("continue");
}
Game.prototype.pause = function () {
  if (!this.iStart || this.iStart == "pause") { return false; }
  if (!!this.pauseTime) { return false; }
  this.pauseTime = new Date().getTime();
  this.tickerStop();
  this.iStart = "pause";
}
// 游戏结束
Game.prototype.gameOver = function (type) {
  if (!this.iStart || this.iStart == "pause" || this.iStart == "over") { return false; }
  this.pauseTime = new Date().getTime();
  this.tickerStop();
  // 防止结束了还在移动
  this.iStart = "over";

  try {
    // 页面销毁时，不调用游戏结束回调
    if (type != "unmount") {
      this.options.overFun({ score: this.totalScore, type });
    }
  } catch (e) { }
}
// 重置游戏
Game.prototype.reset = function () {
  this.tickerStop();
  // 1.隐藏元素
  this.hideSpriteArrFun("spriteArr");

  this.startTime = new Date().getTime();
  // 3.重新生成新元素
  this.init();
}
Game.prototype.startGame = function (type) {
  if (!!this.iStart) { return false; }
  this.iStart = "start";
  this.overSprite = null;
  this.hideSpriteArrFun("animatespriteArr");

  if (type != "continue") {
    this.startTime = new Date().getTime();
  }
  this.tickerStart();
}

export {
  Game
}